# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""Python implementation of higher level PDF constructs."""

from __future__ import annotations

from pikepdf.models._content_stream import (
    ContentStreamInstructions,
    PdfParsingError,  # legacy
    UnparseableContentStreamInstructions,
    parse_content_stream,
    unparse_content_stream,
)
from pikepdf.models.encryption import Encryption, EncryptionInfo, Permissions
from pikepdf.models.image import (
    PdfImage,
    PdfInlineImage,
    UnsupportedImageTypeError,  # legacy
)
from pikepdf.models.metadata import PdfMetadata, XmpDocument
from pikepdf.models.outlines import (
    Outline,
    OutlineItem,
    OutlineStructureError,
    PageLocation,
    make_page_destination,
)

__all__ = [
    'ContentStreamInstructions',
    'PdfParsingError',  # legacy
    'UnparseableContentStreamInstructions',
    'parse_content_stream',
    'unparse_content_stream',
    'Encryption',
    'EncryptionInfo',
    'Permissions',
    'PdfImage',
    'PdfInlineImage',
    'UnsupportedImageTypeError',  # legacy
    'PdfMetadata',
    'XmpDocument',
    'Outline',
    'OutlineItem',
    'OutlineStructureError',  # legacy
    'PageLocation',
    'make_page_destination',
]



# SPDX-FileCopyrightText: 2025 James R. Barlow
# SPDX-License-Identifier: MPL-2.0


from __future__ import annotations

# Conversion mode API
from contextlib import contextmanager
from typing import Literal

from pikepdf import _core


def set_object_conversion_mode(mode: Literal['implicit', 'explicit']) -> None:
    """Set global object conversion mode.

    This controls how PDF scalar values (integers, booleans, reals) are
    returned when accessing PDF objects.

    Args:
        mode: Conversion mode.
            - ``'implicit'`` (default): Automatically convert PDF integers to
              Python ``int``, booleans to ``bool``, and reals to ``Decimal``.
              This is the legacy behavior.
            - ``'explicit'``: Return PDF scalars as ``pikepdf.Integer``,
              ``pikepdf.Boolean``, and ``pikepdf.Real`` objects. This enables
              better type safety and static type checking.

    Example:
        >>> pikepdf.set_object_conversion_mode('explicit')
        >>> pdf = pikepdf.open('test.pdf')
        >>> count = pdf.Root.Count
        >>> isinstance(count, pikepdf.Integer)  # True in explicit mode
        True
        >>> int(count)  # Convert to Python int
        5

    .. versionadded:: 10.1
    """
    _core._set_explicit_conversion_mode(mode == 'explicit')


def get_object_conversion_mode() -> Literal['implicit', 'explicit']:
    """Get current effective object conversion mode.

    This returns the mode that is currently in effect for the calling thread,
    taking into account both the global setting and any active
    :func:`explicit_conversion` context managers.

    Returns:
        The current effective conversion mode: ``'implicit'`` or ``'explicit'``.

    .. versionadded:: 10.1

    .. versionchanged:: 10.2
        Now returns the effective mode, including thread-local context manager state.
    """
    return 'explicit' if _core._get_effective_explicit_mode() else 'implicit'


@contextmanager
def explicit_conversion():
    """Context manager for explicit conversion mode.

    Within this context, PDF scalar values will be returned as
    ``pikepdf.Integer``, ``pikepdf.Boolean``, and ``pikepdf.Real`` objects
    instead of being automatically converted to Python native types.

    This context manager is thread-local: it only affects the current thread
    and takes precedence over the global setting from
    :func:`set_object_conversion_mode`. Nested context managers are supported.

    Example:
        >>> with pikepdf.explicit_conversion():
        ...     pdf = pikepdf.open('test.pdf')
        ...     count = pdf.Root.Count
        ...     isinstance(count, pikepdf.Integer)
        True

    .. versionadded:: 10.1

    .. versionchanged:: 10.2
        Now thread-local and takes precedence over global setting.
    """
    _core._enter_thread_explicit_mode()
    try:
        yield
    finally:
        _core._exit_thread_explicit_mode()



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

"""Defines exceptions that need to be visible to pikepdf._core (C++)."""


class DependencyError(Exception):
    """A third party dependency is needed to extract streams of this type."""



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""XMP metadata constants, templates, and utilities."""

from __future__ import annotations

import re
from collections.abc import Callable, Iterable
from typing import Any, NamedTuple

# XMP Namespace URIs
XMP_NS_DC = "http://purl.org/dc/elements/1.1/"
XMP_NS_PDF = "http://ns.adobe.com/pdf/1.3/"
XMP_NS_PDFA_ID = "http://www.aiim.org/pdfa/ns/id/"
XMP_NS_PDFA_EXTENSION = "http://www.aiim.org/pdfa/ns/extension/"
XMP_NS_PDFA_PROPERTY = "http://www.aiim.org/pdfa/ns/property#"
XMP_NS_PDFA_SCHEMA = "http://www.aiim.org/pdfa/ns/schema#"
XMP_NS_PDFUA_ID = "http://www.aiim.org/pdfua/ns/id/"
XMP_NS_PDFX_ID = "http://www.npes.org/pdfx/ns/id/"
XMP_NS_PHOTOSHOP = "http://ns.adobe.com/photoshop/1.0/"
XMP_NS_PRISM = "http://prismstandard.org/namespaces/basic/1.0/"
XMP_NS_PRISM2 = "http://prismstandard.org/namespaces/basic/2.0/"
XMP_NS_PRISM3 = "http://prismstandard.org/namespaces/basic/3.0/"
XMP_NS_RDF = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
XMP_NS_XMP = "http://ns.adobe.com/xap/1.0/"
XMP_NS_XMP_MM = "http://ns.adobe.com/xap/1.0/mm/"
XMP_NS_XMP_RIGHTS = "http://ns.adobe.com/xap/1.0/rights/"

# This one should not be registered with lxml
XMP_NS_XML = "http://www.w3.org/XML/1998/namespace"

DEFAULT_NAMESPACES: list[tuple[str, str]] = [
    ('adobe:ns:meta/', 'x'),
    (XMP_NS_DC, 'dc'),
    (XMP_NS_PDF, 'pdf'),
    (XMP_NS_PDFA_ID, 'pdfaid'),
    (XMP_NS_PDFA_EXTENSION, 'pdfaExtension'),
    (XMP_NS_PDFA_PROPERTY, 'pdfaProperty'),
    (XMP_NS_PDFA_SCHEMA, 'pdfaSchema'),
    (XMP_NS_PDFUA_ID, 'pdfuaid'),
    (XMP_NS_PDFX_ID, 'pdfxid'),
    (XMP_NS_PHOTOSHOP, 'photoshop'),
    (XMP_NS_PRISM, 'prism'),
    (XMP_NS_PRISM2, 'prism2'),
    (XMP_NS_PRISM3, 'prism3'),
    (XMP_NS_RDF, 'rdf'),
    (XMP_NS_XMP, 'xmp'),
    (XMP_NS_XMP_MM, 'xmpMM'),
    (XMP_NS_XMP_RIGHTS, 'xmpRights'),
    ('http://crossref.org/crossmark/1.0/', 'crossmark'),
    ('http://www.niso.org/schemas/jav/1.0/', 'jav'),
    ('http://ns.adobe.com/pdfx/1.3/', 'pdfx'),
    ('http://www.niso.org/schemas/ali/1.0/', 'ali'),
]


# XMP packet wrappers
XPACKET_BEGIN = b"""<?xpacket begin="\xef\xbb\xbf" id="W5M0MpCehiHzreSzNTczkc9d"?>\n"""

XMP_EMPTY = b"""<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="pikepdf">
 <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
 </rdf:RDF>
</x:xmpmeta>
"""

XPACKET_END = b"""\n<?xpacket end="w"?>\n"""


class XmpContainer(NamedTuple):
    """Map XMP container object to suitable Python container."""

    rdf_type: str
    py_type: type
    insert_fn: Callable[..., None]


class AltList(list):
    """XMP AltList container for language alternatives."""


XMP_CONTAINERS = [
    XmpContainer('Alt', AltList, AltList.append),
    XmpContainer('Bag', set, set.add),
    XmpContainer('Seq', list, list.append),
]


_LANG_ALTS_LAZY = [
    (XMP_NS_DC, 'title'),
    (XMP_NS_DC, 'description'),
    (XMP_NS_DC, 'rights'),
    (XMP_NS_XMP_RIGHTS, 'UsageTerms'),
]

_LOADED_LXML_NAMESPACES = False

# lxml lazy-loading
def __getattr__(name: str) -> Any:
    global _LOADED_LXML_NAMESPACES

    if name == 'LANG_ALTS':
        from lxml.etree import QName

        if not _LOADED_LXML_NAMESPACES:
            from lxml import etree
            # Register all namespaces with lxml
            for _uri, _prefix in DEFAULT_NAMESPACES:
                etree.register_namespace(_prefix, _uri)
            _LOADED_LXML_NAMESPACES = True

        val = frozenset([str(QName(x, y)) for x,y in _LANG_ALTS_LAZY])
        globals()[name] = val

        return val

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


# These are the illegal characters in XML 1.0. (XML 1.1 is a bit more permissive,
# but we'll be strict to ensure wider compatibility.)
re_xml_illegal_chars = re.compile(
    r"(?u)[^\x09\x0A\x0D\x20-\U0000D7FF\U0000E000-\U0000FFFD\U00010000-\U0010FFFF]"
)
re_xml_illegal_bytes = re.compile(rb"[^\x09\x0A\x0D\x20-\xFF]|&#0;")


def clean(s: str | Iterable[str], joiner: str = '; ') -> str:
    """Ensure an object can safely be inserted in a XML tag body.

    If we still have a non-str object at this point, the best option is to
    join it, because it's apparently calling for a new node in a place that
    isn't allowed in the spec or not supported.
    """
    from warnings import warn

    if not isinstance(s, str):
        if isinstance(s, Iterable):
            warn(f"Merging elements of {s}")
            if isinstance(s, set):
                s = joiner.join(sorted(s))
            else:
                s = joiner.join(s)
        else:
            raise TypeError("object must be a string or iterable of strings")
    return re_xml_illegal_chars.sub('', s)



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

from typing import IO, TYPE_CHECKING, Any, AnyStr

if TYPE_CHECKING:
    from lxml.etree import _Element, _ElementTree

_XMLParser = None

def parse_xml(source: AnyStr | IO[Any], recover: bool = False) -> _ElementTree:
    """Wrap lxml's parse to provide protection against XXE attacks."""
    from lxml.etree import XMLParser as _UnsafeXMLParser
    from lxml.etree import parse as _parse

    global _XMLParser

    if _XMLParser is None:
        class _XMLParserImpl(_UnsafeXMLParser):
            def __init__(self, *args: Any, **kwargs: Any):
                # Prevent XXE attacks
                # https://rules.sonarsource.com/python/type/Vulnerability/RSPEC-2755
                kwargs['resolve_entities'] = False
                kwargs['no_network'] = True
                super().__init__(*args, **kwargs)
        _XMLParser = _XMLParserImpl

    parser = _XMLParser(recover=recover, remove_pis=False)
    return _parse(source, parser=parser)

def __getattr__(name: str):
    if name in {'_Element', '_ElementTree'}:
        from lxml import etree
        value = getattr(etree, name)
        globals()[name] = value
        return value

    raise AttributeError(f"module {__name__} has no attribute {name}")

__all__ = ['parse_xml', '_ElementTree', '_Element']



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

"""PDF metadata handling.

This module provides XMP and DocumentInfo metadata access for PDF files.
"""

from __future__ import annotations

from pikepdf.models.metadata._constants import (
    DEFAULT_NAMESPACES,
    XMP_CONTAINERS,
    XMP_EMPTY,
    XMP_NS_DC,
    XMP_NS_PDF,
    XMP_NS_PDFA_EXTENSION,
    XMP_NS_PDFA_ID,
    XMP_NS_PDFA_PROPERTY,
    XMP_NS_PDFA_SCHEMA,
    XMP_NS_PDFUA_ID,
    XMP_NS_PDFX_ID,
    XMP_NS_PHOTOSHOP,
    XMP_NS_PRISM,
    XMP_NS_PRISM2,
    XMP_NS_PRISM3,
    XMP_NS_RDF,
    XMP_NS_XML,
    XMP_NS_XMP,
    XMP_NS_XMP_MM,
    XMP_NS_XMP_RIGHTS,
    XPACKET_BEGIN,
    XPACKET_END,
    AltList,
    XmpContainer,
)
from pikepdf.models.metadata._converters import (
    DOCINFO_MAPPING,
    AuthorConverter,
    Converter,
    DateConverter,
    DocinfoMapping,
    decode_pdf_date,
    encode_pdf_date,
)
from pikepdf.models.metadata._core import PdfMetadata
from pikepdf.models.metadata._xmp import XmpDocument


def __getattr__(name):
    if name == 'LANG_ALTS':
        from pikepdf.models.metadata import _constants
        val = getattr(_constants, 'LANG_ALTS')
        globals()[name] = val
        return val
    raise AttributeError(f"module {__name__} has no attribute {name}")

__all__ = [
    # Main classes
    'PdfMetadata',
    'XmpDocument',
    # Converters
    'Converter',
    'AuthorConverter',
    'DateConverter',
    'DocinfoMapping',
    'DOCINFO_MAPPING',
    'decode_pdf_date',
    'encode_pdf_date',
    # Namespace constants
    'XMP_NS_DC',
    'XMP_NS_PDF',
    'XMP_NS_PDFA_ID',
    'XMP_NS_PDFA_EXTENSION',
    'XMP_NS_PDFA_PROPERTY',
    'XMP_NS_PDFA_SCHEMA',
    'XMP_NS_PDFUA_ID',
    'XMP_NS_PDFX_ID',
    'XMP_NS_PHOTOSHOP',
    'XMP_NS_PRISM',
    'XMP_NS_PRISM2',
    'XMP_NS_PRISM3',
    'XMP_NS_RDF',
    'XMP_NS_XML',
    'XMP_NS_XMP',
    'XMP_NS_XMP_MM',
    'XMP_NS_XMP_RIGHTS',
    # Other exports
    'DEFAULT_NAMESPACES',
    'LANG_ALTS',
    'XPACKET_BEGIN',
    'XPACKET_END',
    'XMP_CONTAINERS',
    'XMP_EMPTY',
    'AltList',
    'XmpContainer',
]
